@@ -1,4 +1,4 @@
-/*	$Id: local2.c,v 1.41 2011/02/18 16:52:37 ragge Exp $	*/
+/*	$Id: local2.c,v 1.58 2014/09/29 19:57:47 ragge Exp $	*/
 /*
  * Copyright (c) 2008 Michael Shalayeff
  * Copyright (c) 2003 Anders Magnusson (ragge@ludd.luth.se).
@@ -42,6 +42,7 @@ deflab(int label)
 static int regoff[MAXREGS];
 static TWORD ftype;
 char *rbyte[], *rshort[], *rlong[];
+static int needframe;
 
 /*
  * Print out the prolog assembler.
@@ -52,7 +53,6 @@ prtprolog(struct interpass_prolog *ipp, int addto)
 {
 	int i;
 
-	/* XXX should look if there is any need to emit this */
 	printf("\tpushq %%rbp\n");
 	printf("\tmovq %%rsp,%%rbp\n");
 	addto = (addto+15) & ~15; /* 16-byte aligned */
@@ -62,7 +62,7 @@ prtprolog(struct interpass_prolog *ipp, int addto)
 	/* save permanent registers */
 	for (i = 0; i < MAXREGS; i++)
 		if (TESTBIT(ipp->ipp_regs, i))
-			fprintf(stdout, "\tmovq %s,-%d(%s)\n",
+			printf("\tmovq %s,-%d(%s)\n",
 			    rnames[i], regoff[i], rnames[FPREG]);
 }
 
@@ -85,6 +85,45 @@ offcalc(struct interpass_prolog *ipp)
 	return addto;
 }
 
+/*
+ * Traverse a tree to check if we need to emit a frame at all.
+ * We emit it if:
+ * - any function call
+ * - rsp or rbp referenced
+ * Return 1 if frame is needed, 0 otherwise.
+ */
+static int
+chkf(NODE *p)
+{
+	int o = p->n_op;
+
+	if ((o == REG || o == OREG) && (regno(p) == RBP || regno(p) == RSP))
+		return 1;
+	if (callop(o))
+		return 1;
+	if (optype(o) == UTYPE)
+		return chkf(p->n_left);
+	else if (optype(o) == BITYPE)
+		return chkf(p->n_left) || chkf(p->n_right);
+	return 0;
+}
+
+static int
+chkframe(struct interpass_prolog *ipp)
+{
+	struct interpass *ip;
+
+	DLIST_FOREACH(ip, &ipp->ipp_ip, qelem) {
+		if (ip->type == IP_EPILOG)
+			break;
+		if (ip->type == IP_NODE) {
+			if (chkf(ip->ip_node))
+				return 1;
+		}
+	}
+	return 0;
+}
+
 void
 prologue(struct interpass_prolog *ipp)
 {
@@ -92,6 +131,11 @@ prologue(struct interpass_prolog *ipp)
 
 	ftype = ipp->ipp_type;
 
+	if (xdeljumps)
+		needframe = chkframe(ipp);
+	else
+		needframe = 1;
+
 #ifdef LANG_F77
 	if (ipp->ipp_vis)
 		printf("	.globl %s\n", ipp->ipp_name);
@@ -103,7 +147,10 @@ prologue(struct interpass_prolog *ipp)
 	 * add to the stack.
 	 */
 	addto = offcalc(ipp);
-	prtprolog(ipp, addto);
+	if (addto)
+		needframe = 1;
+	if (needframe)
+		prtprolog(ipp, addto);
 }
 
 void
@@ -114,21 +161,25 @@ eoftn(struct interpass_prolog *ipp)
 	if (ipp->ipp_ip.ip_lbl == 0)
 		return; /* no code needs to be generated */
 
-	/* return from function code */
-	for (i = 0; i < MAXREGS; i++)
-		if (TESTBIT(ipp->ipp_regs, i))
-			fprintf(stdout, "	movq -%d(%s),%s\n",
-			    regoff[i], rnames[FPREG], rnames[i]);
+	if (needframe) {
+		/* return from function code */
+		for (i = 0; i < MAXREGS; i++)
+			if (TESTBIT(ipp->ipp_regs, i))
+				printf("	movq -%d(%s),%s\n",
+				    regoff[i], rnames[FPREG], rnames[i]);
+
+		/* struct return needs special treatment */
+		if (ftype == STRTY || ftype == UNIONTY) {
+			printf("	movl 8(%%ebp),%%eax\n");
+			printf("	leave\n");
+			printf("	ret $%d\n", 4);
+		} else {
+			printf("	leave\n");
+			printf("	ret\n");
+		}
+	} else
+		printf("\tret\n");
 
-	/* struct return needs special treatment */
-	if (ftype == STRTY || ftype == UNIONTY) {
-		printf("	movl 8(%%ebp),%%eax\n");
-		printf("	leave\n");
-		printf("	ret $%d\n", 4);
-	} else {
-		printf("	leave\n");
-		printf("	ret\n");
-	}
 #ifndef MACHOABI
 	printf("\t.size %s,.-%s\n", ipp->ipp_name, ipp->ipp_name);
 #endif
@@ -212,75 +263,17 @@ fcomp(NODE *p)
 	if (p->n_left->n_op != REG)
 		comperr("bad compare %p\n", p);
 	if ((p->n_su & DORIGHT) == 0)
-		expand(p, 0, "	fxch\n");
-	expand(p, 0, "	fucomip %st(1),%st\n");	/* emit compare insn  */
-	expand(p, 0, "	fstp %st(0)\n");	/* pop fromstack */
+		expand(p, 0, "\tfxch\n");
+	expand(p, 0, "\tfucomip %st(1),%st\n");	/* emit compare insn  */
+	expand(p, 0, "\tfstp %st(0)\n");	/* pop fromstack */
 	zzzcode(p, 'U');
 }
 
 int
 fldexpand(NODE *p, int cookie, char **cp)
 {
-	CONSZ val;
-
-	if (p->n_op == ASSIGN)
-		p = p->n_left;
-	switch (**cp) {
-	case 'S':
-		printf("%d", UPKFSZ(p->n_rval));
-		break;
-	case 'H':
-		printf("%d", UPKFOFF(p->n_rval));
-		break;
-	case 'M':
-	case 'N':
-		val = (((((CONSZ)1 << (UPKFSZ(p->n_rval)-1))-1)<<1)|1);
-		val <<= UPKFOFF(p->n_rval);
-		if (p->n_type > UNSIGNED)
-			printf("0x%llx", (**cp == 'M' ? val : ~val));
-		else
-			printf("0x%llx", (**cp == 'M' ? val : ~val)&0xffffffff);
-		break;
-	default:
-		comperr("fldexpand");
-	}
-	return 1;
-}
-
-static void
-bfext(NODE *p)
-{
-	int ch = 0, sz = 0;
-
-	if (ISUNSIGNED(p->n_right->n_type))
-		return;
-	switch (p->n_right->n_type) {
-	case CHAR:
-		ch = 'b';
-		sz = 8;
-		break;
-	case SHORT:
-		ch = 'w';
-		sz = 16;
-		break;
-	case INT:
-		ch = 'l';
-		sz = 32;
-		break;
-	case LONG:
-		ch = 'q';
-		sz = 64;
-		break;
-	default:
-		comperr("bfext");
-	}
-
-	sz -= UPKFSZ(p->n_left->n_rval);
-	printf("\tshl%c $%d,", ch, sz);
-	adrput(stdout, getlr(p, 'D'));
-	printf("\n\tsar%c $%d,", ch, sz);
-	adrput(stdout, getlr(p, 'D'));
-	printf("\n");
+	comperr("fldexpand");
+	return 0;
 }
 
 static void
@@ -289,7 +282,7 @@ stasg(NODE *p)
 	expand(p, INAREG, "	leaq AL,%rdi\n");
 	if (p->n_stsize >= 8)
 		printf("\tmovl $%d,%%ecx\n\trep movsq\n", p->n_stsize >> 3);
-	if (p->n_stsize & 3)
+	if (p->n_stsize & 4)
 		printf("\tmovsl\n");
 	if (p->n_stsize & 2)
 		printf("\tmovsw\n");
@@ -327,9 +320,6 @@ ultofd(NODE *p)
 static void
 ldtoul(NODE *p)
 {
-	int r;
-
-	r = getlr(p, '1')->n_rval;
 
 	E("	subq $16,%rsp\n");
 	E("	movl $0x5f000000,(%rsp)\n"); /* More than long can have */
@@ -420,10 +410,37 @@ zzzcode(NODE *p, int c)
 			return; /* XXX remove ZC from UCALL */
 		if (pr)
 			printf("	addq $%d, %s\n", pr, rnames[RSP]);
-		break;
-
-	case 'E': /* Perform bitfield sign-extension */
-		bfext(p);
+#define	STRREG 6
+#define	STRSSE 8
+#define	STRIF  9
+#define	STRFI  10
+#define	STRX87 11
+		if ((p->n_op == STCALL || p->n_op == USTCALL) &&
+		    p->n_stsize == 32 && p->n_stalign == STRX87) {
+			printf("\tfstpt -%d(%%rbp)\n", stkpos);
+			printf("\tfstpt -%d(%%rbp)\n", stkpos-16);
+			printf("\tleaq -%d(%%rbp),%%rax\n", stkpos);
+		}
+		if ((p->n_op == STCALL || p->n_op == USTCALL) &&
+		    p->n_stsize <= 16) {
+			/* store reg-passed structs on stack */
+			if (p->n_stalign == STRREG || p->n_stalign == STRIF)
+				printf("\tmovq %%rax,-%d(%%rbp)\n", stkpos);
+			else
+				printf("\tmovsd %%xmm0,-%d(%%rbp)\n", stkpos);
+			if (p->n_stsize > 8) {
+				if (p->n_stalign == STRREG)
+					printf("\tmovq %%rdx");
+				else if (p->n_stalign == STRFI)
+					printf("\tmovq %%rax");
+				else if (p->n_stalign == STRIF)
+					printf("\tmovsd %%xmm0");
+				else
+					printf("\tmovsd %%xmm1");
+				printf(",-%d(%%rbp)\n", stkpos-8);
+			}
+			printf("\tleaq -%d(%%rbp),%%rax\n", stkpos);
+		}
 		break;
 
 	case 'F': /* Structure argument */
@@ -456,7 +473,8 @@ zzzcode(NODE *p, int c)
 		break;
 
 	case 'P': /* Put hidden argument in rdi */
-		printf("\tleaq -%d(%%rbp),%%rdi\n", stkpos);
+		if (p->n_stsize > 16 && p->n_stalign != STRX87)
+			printf("\tleaq -%d(%%rbp),%%rdi\n", stkpos);
 		break;
 
         case 'Q': /* emit struct assign */
@@ -512,13 +530,6 @@ zzzcode(NODE *p, int c)
 	}
 }
 
-/*ARGSUSED*/
-int
-rewfld(NODE *p)
-{
-	return(1);
-}
-
 int canaddr(NODE *);
 int
 canaddr(NODE *p)
@@ -537,13 +548,8 @@ canaddr(NODE *p)
 int
 flshape(NODE *p)
 {
-	int o = p->n_op;
-
-	if (o == OREG || o == REG || o == NAME)
-		return SRDIR; /* Direct match */
-	if (o == UMUL && shumul(p->n_left, SOREG))
-		return SROREG; /* Convert into oreg */
-	return SRREG; /* put it into a register */
+	comperr("flshape");
+	return(0);
 }
 
 /* INTEMP shapes must not contain any temporary registers */
@@ -627,7 +633,7 @@ upput(NODE *p, int size)
 	size /= SZCHAR;
 	switch (p->n_op) {
 	case REG:
-		fprintf(stdout, "%%%s", &rnames[p->n_rval][3]);
+		printf("%%%s", &rnames[p->n_rval][3]);
 		break;
 
 	case NAME:
@@ -637,7 +643,7 @@ upput(NODE *p, int size)
 		p->n_lval -= size;
 		break;
 	case ICON:
-		fprintf(stdout, "$" CONFMT, p->n_lval >> 32);
+		printf("$" CONFMT, p->n_lval >> 32);
 		break;
 	default:
 		comperr("upput bad op %d size %d", p->n_op, size);
@@ -651,9 +657,6 @@ adrput(FILE *io, NODE *p)
 	char **rc;
 	/* output an address, with offsets, from p */
 
-	if (p->n_op == FLD)
-		p = p->n_left;
-
 	switch (p->n_op) {
 
 	case NAME:
@@ -683,18 +686,6 @@ adrput(FILE *io, NODE *p)
 			fprintf(io, "(%s)", rnames[p->n_rval]);
 		return;
 	case ICON:
-#ifdef PCC_DEBUG
-		/* Sanitycheck for PIC, to catch adressable constants */
-		if (kflag && p->n_name[0]) {
-			static int foo;
-
-			if (foo++ == 0) {
-				printf("\nfailing...\n");
-				fwalk(p, e2print, 0);
-				comperr("pass2 conput");
-			}
-		}
-#endif
 		/* addressable value of the constant */
 		fputc('$', io);
 		conput(io, p);
@@ -752,15 +743,52 @@ cbgen(int o, int lab)
 	printf("	%s " LABFMT "\n", ccbranches[o-EQ], lab);
 }
 
+/*
+ * gcc xasm has the ability to generate different asm types
+ * via some magic.
+ *
+ * Only support AT&T asm for now.
+ */
+static char *
+adjustname(char *s)
+{
+	int len = strlen(s);
+	char *d = tmpalloc(len+1);
+	int i, j, flvl, tlvl;
+
+	flvl = tlvl = 0;
+	for (i = j = 0; i < len; i++) {
+		switch (s[i]) {
+		case '{': tlvl++; break;
+		case '}': if (tlvl)tlvl--; else flvl--; break;
+		case '|': tlvl--; flvl++; break;
+		default:
+			if (flvl == 0)
+				d[j++] = s[i];
+			break;
+		}
+	}
+	d[j] = 0;
+	return d;
+}
+
 static void
 fixcalls(NODE *p, void *arg)
 {
+	int ps;
+
 	/* Prepare for struct return by allocating bounce space on stack */
 	switch (p->n_op) {
 	case STCALL:
 	case USTCALL:
-		if (p->n_stsize+p2autooff > stkpos)
-			stkpos = p->n_stsize+p2autooff;
+		ps = p->n_stsize;
+		if (ps < 16)
+			ps = 16;
+		if (ps+p2autooff > stkpos)
+			stkpos = ps+p2autooff;
+		break;
+	case XASM:
+		p->n_name = adjustname(p->n_name);
 		break;
 	}
 }
@@ -1027,6 +1055,8 @@ retry:	switch (c) {
 	case 'c': reg = RCX; break;
 	case 'd': reg = RDX; break;
 
+	case 'Q': reg = RDX; break; /* Always dx for now */
+
 	case 'x':
 	case 'q':
 	case 't':
@@ -1034,7 +1064,7 @@ retry:	switch (c) {
 		p->n_name = tmpstrdup(p->n_name);
 		w = strchr(p->n_name, c);
 		*w = 'r'; /* now reg */
-		return c == 'q' || c == 'x' ? 0 : 1;
+		return c == 'q' || c == 'x' || c == 't' ? 0 : 1;
 
 	case 'I':
 	case 'J':
@@ -1043,8 +1073,13 @@ retry:	switch (c) {
 	case 'M':
 	case 'N':
 		if (p->n_left->n_op != ICON) {
-			if ((c = XASMVAL1(cw)))
+			if ((c = XASMVAL1(cw))) {
+				if (c == 'r') {
+					p->n_name++;
+					return 0;
+				}
 				goto retry;
+			}
 			uerror("xasm arg not constant");
 		}
 		v = p->n_left->n_lval;
@@ -1112,6 +1147,13 @@ targarg(char *w, void *arg, int n)
 	if (q->n_op == REG) {
 		if (*w == 'k') {
 			q->n_type = INT;
+		} else if (*w == 'q') {
+			q->n_type = LONG;
+		} else if (*w == 'h' || *w == 'b') {
+			/* Can do this only because we know dx is used */
+			printf("%%d%c", *w == 'h' ? 'h' : 'l');
+			tfree(q);
+			return;
 		} else if (*w != 'w') {
 			cerror("targarg"); /* XXX ??? */
 			if (q->n_type > UCHAR) {
@@ -1158,6 +1200,14 @@ static struct {
 	{ "rdx", RDX },
 	{ "rsi", RSI },
 	{ "rdi", RDI },
+	{ "r8", R08 },
+	{ "r9", R09 },
+	{ "r10", R10 },
+	{ "r11", R11 },
+	{ "r12", R12 },
+	{ "r13", R13 },
+	{ "r14", R14 },
+	{ "r15", R15 },
 	{ "st", 040 },
 	{ "st(0)", 040 },
 	{ "st(1)", 041 },
